""" Common tools for tsrc commands """

import os

import path

import tsrc
import tsrc.workspace


def find_workspace_path():
    """ Look for a workspace root somewhere in the upper directories
    hierarchy

    """
    head = os.getcwd()
    tail = True
    while tail:
        tsrc_path = os.path.join(head, ".tsrc")
        if os.path.isdir(tsrc_path):
            return path.Path(head)
        tbuild_yml_path = os.path.join(head, "tbuild.yml")
        if os.path.exists(tbuild_yml_path):
            return path.Path(head)

        else:
            head, tail = os.path.split(head)
    raise tsrc.Error("Could not find current workspace")


def get_workspace(args):
    if args.workspace_path:
        workspace_path = path.Path(args.workspace_path)
    else:
        workspace_path = find_workspace_path()
    return tsrc.workspace.Workspace(workspace_path)



import os

import pytest

import tsrc.cli

from tsrc.test.conftest import *


def get_cmd_for_foreach_test(shell=False):
    """ We need a cmd that:
     * can fail if not called with the correct 'shell'
       argument
       (to check `foreach -c` option)
     * can fail if a directory does not exist
       (to check `foreach` error handling)
    """
    if os.name == "nt":
        if shell:
            cmd = ["dir"]
        else:
            cmd = ["cmd.exe", "/c", "dir"]
    else:
        if shell:
            cmd = ["cd"]
        else:
            cmd = ["ls"]
    return cmd


def test_foreach_no_args(tsrc_cli, git_server):
    git_server.add_repo("foo")
    tsrc_cli.run("init", git_server.manifest_url)
    tsrc_cli.run("foreach", expect_fail=True)


def test_foreach_with_errors(tsrc_cli, git_server, message_recorder):
    git_server.add_repo("foo")
    git_server.add_repo("spam")
    git_server.push_file("foo", "foo/bar.txt",
                         contents="this is bar")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    cmd = get_cmd_for_foreach_test(shell=False)
    cmd.append("foo")
    tsrc_cli.run("foreach", *cmd, expect_fail=True)
    assert message_recorder.find("Running `.*` .* failed")
    assert message_recorder.find("\* spam")


def test_foreach_happy(tsrc_cli, git_server, message_recorder):
    git_server.add_repo("foo")
    git_server.add_repo("spam")
    git_server.push_file("foo", "doc/index.html")
    git_server.push_file("spam", "doc/index.html")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    cmd = get_cmd_for_foreach_test(shell=False)
    cmd.append("doc")
    tsrc_cli.run("foreach", *cmd)
    assert message_recorder.find("`%s`" % " ".join(cmd))


def test_foreach_shell(tsrc_cli, git_server, message_recorder):
    git_server.add_repo("foo")
    git_server.add_repo("spam")
    git_server.push_file("foo", "doc/index.html")
    git_server.push_file("spam", "doc/index.html")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    cmd = get_cmd_for_foreach_test(shell=True)
    cmd.append("doc")
    tsrc_cli.run("foreach", "-c", " ".join(cmd))
    assert message_recorder.find("`%s`" % " ".join(cmd))



import textwrap

import schema

import tsrc.config

import pytest
import mock


def test_read_config(tmp_path):
    tsrc_yml_path = tmp_path.joinpath("tsrc.yml")
    tsrc_yml_path.write_text(
        textwrap.dedent(
            """\
            auth:
              gitlab:
                token: MY_SECRET_TOKEN
            """)
    )
    config = tsrc.config.parse_tsrc_config(config_path=tsrc_yml_path)
    assert config["auth"]["gitlab"]["token"] == "MY_SECRET_TOKEN"


def test_invalid_syntax(tmp_path):
    foo_yml = tmp_path.joinpath("foo.yml")
    foo_yml.write_text(textwrap.dedent(
        """
        foo:
          bar:
            baz: [

        baz: 42
        """))
    with pytest.raises(tsrc.InvalidConfig) as e:
        dummy_schema = mock.Mock()
        tsrc.config.parse_config_file(foo_yml, dummy_schema)
    assert e.value.path == foo_yml
    assert "flow sequence" in e.value.details
    assert "ligne 3, col 9" in e.value.details


def test_invalid_schema(tmp_path):
    foo_yml = tmp_path.joinpath("foo.yml")
    foo_yml.write_text(textwrap.dedent(
        """
        foo:
            bar: 42
        """
    ))
    foo_schema = schema.Schema(
        {"foo": {"bar": str}}
    )
    with pytest.raises(tsrc.InvalidConfig) as e:
        tsrc.config.parse_config_file(foo_yml, foo_schema)
    assert "42 should be instance of 'str'" in e.value.details



""" Support for finding elements inside a list of groups """

import tsrc


class GroupError(tsrc.Error):
    pass


class GroupNotFound(GroupError):
    def __init__(self, group_name, parent_group=None):
        self.group_name = group_name
        self.parent_group = parent_group
        if self.parent_group:
            message = "Invalid include detected for %s:\n" % self.parent_group.name
        else:
            message = ""
        message += "No such group: %s" % self.group_name
        super().__init__(message)


class UnknownElement(GroupError):
    def __init__(self, group_name, element):
        self.group_name = group_name
        self.element = element
        message = "%s: unknown element: %s" % (group_name, element)
        super().__init__(message)


# pylint: disable=too-few-public-methods
class Group:
    def __init__(self, name, elements, includes=None):
        self.name = name
        self.elements = elements
        self.includes = includes or list()


class GroupList:
    def __init__(self, *, elements):
        self.groups = dict()
        self.all_elements = elements
        self._groups_seen = set()

    def add(self, name, elements, includes=None):
        for element in elements:
            if element not in self.all_elements:
                raise UnknownElement(name, element)
        self.groups[name] = Group(name, elements, includes=includes)

    def get_group(self, name):
        return self.groups.get(name)

    def get_elements(self, groups=None):
        self._groups_seen = set()
        res = set()
        if not groups:
            return self.all_elements
        self._rec_get_elements(res, groups, parent_group=None)
        return res

    def _rec_get_elements(self, res, group_names, *, parent_group):
        for group_name in group_names:
            if group_name in self._groups_seen:
                return
            if group_name not in self.groups:
                raise GroupNotFound(group_name, parent_group=parent_group)
            group = self.groups[group_name]
            for element in group.elements:
                res.add(element)
            self._groups_seen.add(group.name)
            self._rec_get_elements(res, group.includes, parent_group=group)



""" Tiny wrapper for gitlab REST API """

import urllib.parse

import requests
import ui

import tsrc

GITLAB_API_VERSION = "v4"


class GitLabError(tsrc.Error):
    pass


class GitLabAPIError(GitLabError):
    def __init__(self, url, status_code, message):
        super().__init__(message)
        self.url = url
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return "%s - %s" % (self.status_code, self.message)


class TooManyUsers(GitLabError):
    pass


def handle_errors(response, stream=False):
    if stream:
        _handle_stream_errors(response)
    else:
        _handle_json_errors(response)


def _handle_json_errors(response):
    # Make sure we always have a dict containing some
    # kind of error:
    json_details = dict()
    try:
        json_details = response.json()
    except ValueError:
        json_details["error"] = ("Expecting json result, got %s" % response.text)

    status_code = response.status_code
    url = response.url
    if 400 <= status_code < 500:
        for key in ["error", "message"]:
            if key in json_details:
                raise GitLabAPIError(url, status_code, json_details[key])
        raise GitLabAPIError(url, status_code, json_details)
    if status_code >= 500:
        raise GitLabAPIError(url, status_code, response.text)


def _handle_stream_errors(response):
    if response.status_code >= 400:
        raise GitLabAPIError(response.url, "Incorrect status code:", response.status_code)


class GitLabHelper():
    def __init__(self, gitlab_url, token):
        self.gitlab_api_url = gitlab_url + "/api/" + GITLAB_API_VERSION
        self.token = token

    def make_request(self, verb, url, *, data=None, params=None, stream=False):
        response = self.get_response(verb, url, data=data, params=params, stream=stream)
        handle_errors(response, stream=stream)
        if stream:
            return response
        else:
            return response.json()

    def get_response(self, verb, url, *, data=None, params=None, stream=False):
        full_url = self.gitlab_api_url + url
        response = requests.request(verb, full_url,
                                    headers={"PRIVATE-TOKEN": self.token},
                                    data=data, params=params, stream=stream)
        return response

    def get_project_id(self, project_name):
        encoded_project_name = urllib.parse.quote(project_name, safe=list())
        url = "/projects/%s" % encoded_project_name
        try:
            res = self.make_request("GET", url)
            return res["id"]
        except GitLabAPIError as e:
            if e.status_code == 404:
                raise GitLabAPIError(url, 404, "Project not found: %s" % project_name) from None
            else:
                raise

    def find_opened_merge_request(self, project_id, source_branch):
        url = "/projects/%s/merge_requests" % project_id
        previous_mrs = self.make_request("GET", url)
        for mr in previous_mrs:
            if mr["source_branch"] == source_branch:
                if mr["state"] == "opened":
                    return mr

    def create_merge_request(self, project_id, source_branch, *, title,
                             target_branch="master"):
        ui.info_2("Creating merge request", ui.ellipsis, end="")
        url = "/projects/%i/merge_requests" % project_id
        data = {
            "source_branch": source_branch,
            "target_branch": target_branch,
            "title": title,
            "project_id": project_id,
        }
        result = self.make_request("POST", url, data=data)
        ui.info("done", ui.check)
        return result

    def update_merge_request(self, merge_request, **kwargs):
        project_id = merge_request["target_project_id"]
        merge_request_iid = merge_request["iid"]
        url = "/projects/%s/merge_requests/%s" % (project_id, merge_request_iid)
        return self.make_request("PUT", url, data=kwargs)

    def accept_merge_request(self, merge_request):
        project_id = merge_request["project_id"]
        merge_request_iid = merge_request["iid"]
        ui.info_2("Merging when build succeeds", ui.ellipsis, end="")
        url = "/projects/%s/merge_requests/%s/merge" % (project_id, merge_request_iid)
        data = {
            "merge_when_pipeline_succeeds": True,
        }
        self.make_request("PUT", url, data=data)
        ui.info("done", ui.check)

    def get_active_users(self):
        response = self.get_response("GET", "/users", params={"active": "true", "per_page": 100})
        total = int(response.headers["X-TOTAL"])
        if total > 100:
            raise TooManyUsers()
        else:
            return response.json()

    def get_group_members(self, group, query=None):
        return self.make_request("GET", "/groups/%s/members" % group,
                                 params={"query": query})

    def get_project_members(self, project_id, query=None):
        return self.make_request("GET", "/projects/%s/members" % project_id,
                                 params={"query": query})



""" manifests for tsrc """

import operator
import os

import schema

import tsrc
import tsrc.config
import tsrc.groups


class RepoNotFound(tsrc.Error):
    def __init__(self, src):
        super().__init__("No repo found in '%s'" % src)


def load(manifest_path):
    gitlab_schema = {"url": str}
    copy_schema = {"src": str, schema.Optional("dest"): str}
    repo_schema = {
        "src": str,
        "url": str,
        schema.Optional("branch"): str,
        schema.Optional("copy"): [copy_schema],
        schema.Optional("fixed_ref"): str,
    }
    group_schema = {
        "repos": [str],
        schema.Optional("includes"): [str],
    }
    manifest_schema = schema.Schema({
        "repos": [repo_schema],
        schema.Optional("gitlab"): gitlab_schema,
        schema.Optional("groups"): {str: group_schema},
    })
    parsed = tsrc.config.parse_config_file(manifest_path, manifest_schema)
    res = Manifest()
    res.load(parsed)
    return res


class Manifest():
    def __init__(self):
        self._repos = list()
        self.copyfiles = list()
        self.gitlab = dict()
        self.group_list = None

    def load(self, config):
        self.copyfiles = list()
        self.gitlab = config.get("gitlab")
        repos = config.get("repos") or list()
        for repo_config in repos:
            url = repo_config["url"]
            src = repo_config["src"]
            branch = repo_config.get("branch", "master")
            fixed_ref = repo_config.get("fixed_ref")
            repo = tsrc.Repo(url=url, src=src, branch=branch,
                             fixed_ref=fixed_ref)
            self._repos.append(repo)

            self._handle_copies(repo_config)

        self._handle_groups(config)

    def _handle_copies(self, repo_config):
        if "copy" not in repo_config:
            return
        to_cp = repo_config["copy"]
        for item in to_cp:
            src_copy = item["src"]
            dest_copy = item.get("dest", src_copy)
            src_copy = os.path.join(repo_config["src"], src_copy)
            self.copyfiles.append((src_copy, dest_copy))

    def _handle_groups(self, config):
        elements = set([repo.src for repo in self._repos])
        self.group_list = tsrc.groups.GroupList(elements=elements)
        groups_config = config.get("groups", dict())
        for name, group_config in groups_config.items():
            elements = set(group_config["repos"])
            includes = set(group_config.get("includes", list()))
            self.group_list.add(name, elements, includes=includes)

    def get_repos(self, groups=None, all_=False):
        if all_:
            return self._repos

        if not groups:
            if self._has_default_group():
                return self._get_repos_in_groups(["default"])
            else:
                return self._repos

        return self._get_repos_in_groups(groups)

    def _has_default_group(self):
        return self.group_list.get_group("default") is not None

    def _get_repos_in_groups(self, groups):
        elements = self.group_list.get_elements(groups=groups)
        res = list()
        for src in elements:
            res.append(self.get_repo(src))
        return sorted(res, key=operator.attrgetter("src"))

    def get_url(self, src):
        repo = self.get_repo(src)
        return repo.url

    def get_repo(self, src):
        for repo in self._repos:
            if repo.src == src:
                return repo
        raise RepoNotFound(src)



""" Entry point for tsrc push """

import re

import ui

import tsrc
import tsrc.config
import tsrc.gitlab
import tsrc.git
import tsrc.cli


WIP_PREFIX = "WIP: "


class NoUserMatching(tsrc.Error):
    def __init__(self, query):
        self.query = query
        super().__init__("No user found matching: %s" % self.query)


def get_token():
    config = tsrc.config.parse_tsrc_config()
    return config["auth"]["gitlab"]["token"]


def get_project_name(repo_path):
    rc, out = tsrc.git.run_git(repo_path, "remote", "get-url", "origin", raises=False)
    if rc != 0:
        ui.fatal("Could not get url of 'origin' remote:", out)
    repo_url = out
    return project_name_from_url(repo_url)


def project_name_from_url(url):
    """
    >>> project_name_from_url('git@example.com:foo/bar.git')
    'foo/bar'
    >>> project_name_from_url('ssh://git@example.com:8022/foo/bar.git')
    'foo/bar'
    """
    # split everthing that is separated by a colon or a slash
    parts = re.split("[:/]", url)
    # join the last two parts
    res = "/".join(parts[-2:])
    # remove last `.git`
    if res.endswith(".git"):
        res = res[:-4]
    return res


def select_assignee(choices):
    return ui.ask_choice("Select an assignee", choices, func_desc=lambda x: x["name"])


def wipify(title):
    if not title.startswith(WIP_PREFIX):
        return WIP_PREFIX + title


def unwipify(title):
    if title.startswith(WIP_PREFIX):
        return title[len(WIP_PREFIX):]


class PushAction():
    def __init__(self, args, gl_helper=None):
        self.args = args
        self.gl_helper = gl_helper
        self.source_branch = None
        self.target_branch = None
        self.project_id = None
        self.project_name = None
        self.repo_path = None
        self.source_branch = None
        self.target_branch = None

    def main(self):
        self.prepare()
        self.push()
        self.handle_merge_request()

    def prepare(self):
        if not self.gl_helper:
            workspace = tsrc.cli.get_workspace(self.args)
            workspace.load_manifest()
            gitlab_url = workspace.get_gitlab_url()
            token = get_token()
            self.gl_helper = tsrc.gitlab.GitLabHelper(gitlab_url, token)

        self.repo_path = tsrc.git.get_repo_root()
        self.project_name = get_project_name(self.repo_path)
        self.project_id = self.gl_helper.get_project_id(self.project_name)

        current_branch = tsrc.git.get_current_branch(self.repo_path)
        self.source_branch = current_branch
        self.target_branch = self.args.target_branch

    def push(self):
        ui.info_2("Running git push")
        cmd = ["push", "-u", "origin", "%s:%s" % (self.source_branch, self.source_branch)]
        if self.args.force:
            cmd.append("--force")
        tsrc.git.run_git(self.repo_path, *cmd)

    def handle_assignee(self):
        if not self.args.assignee:
            return None
        return self.find_assigne(self.args.assignee)

    def get_review_candidates(self, query):
        group_name = self.project_name.split("/")[0]
        project_members = self.gl_helper.get_project_members(self.project_id, query=query)
        group_members = self.gl_helper.get_group_members(group_name, query=query)
        # Concatenate and de-duplicate results:
        candidates = project_members + group_members
        res = list()
        seen = set()
        for user in candidates:
            user_name = user["name"]
            if user_name not in seen:
                seen.add(user_name)
                res.append(user)
        return res

    def find_assigne(self, query):
        candidates = self.get_review_candidates(query=query)
        if not candidates:
            raise NoUserMatching(query)
        if len(candidates) == 1:
            return candidates[0]
        return select_assignee(candidates)

    def handle_merge_request(self):
        assignee = self.handle_assignee()
        if assignee:
            ui.info_2("Assigning to", assignee["name"])

        merge_request = self.ensure_merge_request()

        title = self.handle_title(merge_request)

        params = {
            "title": title,
            "target_branch": self.target_branch,
            "remove_source_branch": True,
        }
        if assignee:
            params["assignee_id"] = assignee["id"]

        self.gl_helper.update_merge_request(merge_request, **params)

        if self.args.accept:
            self.gl_helper.accept_merge_request(merge_request)

        ui.info(ui.green, "::",
                ui.reset, "See merge request at", merge_request["web_url"])

    def handle_title(self, merge_request):
        # If set from command line: use it
        if self.args.mr_title:
            return self.args.mr_title
        else:
            # Change the title if we need to
            title = merge_request["title"]
            if self.args.ready:
                return unwipify(title)
            if self.args.wip:
                return wipify(title)

    def find_merge_request(self):
        return self.gl_helper.find_opened_merge_request(
            self.project_id, self.source_branch
        )

    def create_merge_request(self):
        return self.gl_helper.create_merge_request(
            self.project_id, self.source_branch,
            title=self.source_branch,
            target_branch=self.target_branch
        )

    def ensure_merge_request(self):
        merge_request = self.find_merge_request()
        if merge_request:
            ui.info_2("Found existing merge request: !%s" % merge_request["iid"])
            return merge_request
        else:
            return self.create_merge_request()


def main(args):
    push_action = PushAction(args)
    push_action.main()



""" Entry point for tsrc foreach """

import subprocess

import ui

import tsrc.cli
import tsrc.workspace


class CommandFailed(tsrc.Error):
    pass


class CmdRunner(tsrc.executor.Task):
    def __init__(self, workspace, cmd, cmd_as_str, shell=False):
        self.workspace = workspace
        self.cmd = cmd
        self.cmd_as_str = cmd_as_str
        self.shell = shell

    def display_item(self, repo):
        return repo.src

    def description(self):
        return "Running `%s` on every repo" % self.cmd_as_str

    def process(self, repo):
        ui.info(repo.src, "\n",
                ui.lightgray, "$ ",
                ui.reset, ui.bold, self.cmd_as_str,
                sep="")
        full_path = self.workspace.joinpath(repo.src)
        rc = subprocess.call(self.cmd, cwd=full_path, shell=self.shell)
        if rc != 0:
            raise CommandFailed()


def main(args):
    workspace = tsrc.cli.get_workspace(args)
    workspace.load_manifest()
    cmd_runner = CmdRunner(workspace, args.cmd, args.cmd_as_str, shell=args.shell)
    tsrc.executor.run_sequence(workspace.get_repos(), cmd_runner)
    ui.info("OK", ui.check)



import os

import pytest

import tsrc.cli

from tsrc.test.conftest import *


def test_sync_happy(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("foo/bar")
    git_server.add_repo("spam/eggs")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    git_server.push_file("foo/bar", "bar.txt", contents="this is bar")

    tsrc_cli.run("sync")

    bar_txt_path = workspace_path.joinpath("foo", "bar", "bar.txt")
    assert bar_txt_path.text() == "this is bar"


def test_sync_with_errors(tsrc_cli, git_server, workspace_path, message_recorder):
    git_server.add_repo("foo/bar")
    git_server.add_repo("spam/eggs")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    git_server.push_file("foo/bar", "bar.txt", contents="Bar is true")
    bar_src = workspace_path.joinpath("foo/bar")
    bar_src.joinpath("bar.txt").write_text("Bar is false")

    tsrc_cli.run("sync", expect_fail=True)

    assert message_recorder.find("Synchronize workspace failed")
    assert message_recorder.find("\* foo/bar")


def test_sync_finds_root(tsrc_cli, git_server, workspace_path, monkeypatch):
    git_server.add_repo("foo/bar")
    tsrc_cli.run("init", git_server.manifest_url)
    monkeypatch.chdir(workspace_path.joinpath("foo/bar"))
    tsrc_cli.run("sync")


def test_new_repo_added_to_manifest(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("foo/bar")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    git_server.add_repo("spam/eggs")

    tsrc_cli.run("sync")

    assert workspace_path.joinpath("spam/eggs").exists()


def test_switching_manifest_branches(tsrc_cli, git_server, workspace_path):
    # Init with manifest_url on master
    git_server.add_repo("foo")
    tsrc_cli.run("init", git_server.manifest_url)

    # Create a new repo, bar, but only on the 'devel'
    # branch of the manifest
    git_server.manifest.change_branch("devel")
    git_server.add_repo("bar")
    bar_path = workspace_path.joinpath("bar")

    # Sync on master branch: bar should not be cloned
    tsrc_cli.run("sync")
    assert not bar_path.exists()

    # Re-init with --branch=devel, bar should be cloned
    tsrc_cli.run("init", "--branch", "devel", git_server.manifest_url)
    tsrc_cli.run("sync")
    assert bar_path.exists()


def test_sync_not_on_master(tsrc_cli, git_server, workspace_path, message_recorder):
    git_server.add_repo("foo")
    git_server.add_repo("bar")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    foo_path = workspace_path.joinpath("foo")
    tsrc.git.run_git(foo_path, "checkout", "-B", "devel")
    # push so that sync still works
    tsrc.git.run_git(foo_path, "push", "-u", "origin", "devel", "--no-verify")

    tsrc_cli.run("sync", expect_fail=True)

    assert message_recorder.find("not on the correct branch")


def test_copies_are_up_to_date(tsrc_cli, git_server, workspace_path):
    manifest_url = git_server.manifest_url
    git_server.add_repo("foo")
    git_server.push_file("foo", "foo.txt", contents="v1")
    git_server.manifest.set_repo_file_copies("foo", [["foo.txt", "top.txt"]])
    tsrc_cli.run("init", manifest_url)
    git_server.push_file("foo", "foo.txt", contents="v2")

    tsrc_cli.run("sync")

    assert workspace_path.joinpath("top.txt").text() == "v2"


def test_copies_are_readonly(tsrc_cli, git_server, workspace_path):
    manifest_url = git_server.manifest_url
    git_server.add_repo("foo")
    git_server.push_file("foo", "foo.txt", contents="v1")
    git_server.manifest.set_repo_file_copies("foo", [["foo.txt", "top.txt"]])

    tsrc_cli.run("init", manifest_url)

    foo_txt = workspace_path.joinpath("top.txt")
    assert not os.access(foo_txt, os.W_OK)


def test_changing_branch(tsrc_cli, git_server, workspace_path, message_recorder):
    git_server.add_repo("foo")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)

    git_server.change_repo_branch("foo", "next")
    git_server.push_file("foo", "next.txt")
    git_server.manifest.set_repo_branch("foo", "next")

    tsrc_cli.run("sync", expect_fail=True)
    assert message_recorder.find("not on the correct branch")


def test_fixed_ref_are_not_updated(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("foo")
    git_server.tag("foo", "v0.1")
    git_server.manifest.set_repo_ref("foo", "v0.1")

    tsrc_cli.run("init", git_server.manifest_url)

    git_server.push_file("foo", "new.txt")

    tsrc_cli.run("sync")

    foo_path = workspace_path.joinpath("foo")
    assert not foo_path.joinpath("new.txt").exists()


def test_fixed_ref_are_updated_when_clean(tsrc_cli, git_server, workspace_path):

    git_server.add_repo("foo")
    git_server.tag("foo", "v0.1")
    git_server.manifest.set_repo_ref("foo", "v0.1")

    tsrc_cli.run("init", git_server.manifest_url)

    git_server.push_file("foo", "new.txt")
    git_server.tag("foo", "v0.2")
    git_server.manifest.set_repo_ref("foo", "v0.2")

    tsrc_cli.run("sync")

    foo_path = workspace_path.joinpath("foo")
    assert foo_path.joinpath("new.txt").exists()


def test_fixed_ref_are_skipped_when_not_clean(tsrc_cli, git_server, workspace_path):

    git_server.add_repo("foo")
    git_server.tag("foo", "v0.1")
    git_server.manifest.set_repo_ref("foo", "v0.1")

    tsrc_cli.run("init", git_server.manifest_url)
    workspace_path.joinpath("foo", "untracked.txt").write_text("")

    git_server.push_file("foo", "new.txt")
    git_server.tag("foo", "v0.2")
    git_server.manifest.set_repo_ref("foo", "v0.2")

    tsrc_cli.run("sync", expect_fail=True)

    foo_path = workspace_path.joinpath("foo")
    assert not foo_path.joinpath("new.txt").exists()


def test_custom_group(tsrc_cli, git_server, message_recorder):
    git_server.add_group("foo", ["bar", "baz"])
    git_server.add_repo("other")

    tsrc_cli.run("init", git_server.manifest_url, "--group", "foo")

    message_recorder.reset()
    tsrc_cli.run("sync")
    assert message_recorder.find("bar")
    assert not message_recorder.find("other")



""" git tools """


import os
import subprocess

import path
import ui

import tsrc


class GitError(tsrc.Error):
    pass


class GitCommandError(GitError):
    def __init__(self, working_path, cmd, *, output=None):
        self.cmd = cmd
        self.working_path = working_path
        self.output = output
        message = "`git {cmd}` from {working_path} failed"
        message = message.format(cmd=" ".join(cmd), working_path=working_path)
        if output:
            message += "\n" + output
        super().__init__(message)


class WorktreeNotFound(GitError):
    def __init__(self, working_path):
        super().__init__("'{}' is not inside a git repository".format(working_path))


def run_git(working_path, *cmd, raises=True):
    """ Run git `cmd` in given `working_path`

    If `raises` is True and git return code is non zero, raise
    an exception. Otherwise, return a tuple (returncode, out)

    """
    git_cmd = list(cmd)
    git_cmd.insert(0, "git")
    options = dict()
    if not raises:
        options["stdout"] = subprocess.PIPE
        options["stderr"] = subprocess.STDOUT

    ui.debug(ui.lightgray, working_path, "$", ui.reset, *git_cmd)
    process = subprocess.Popen(git_cmd, cwd=working_path, **options)

    if raises:
        process.wait()
    else:
        out, _ = process.communicate()
        out = out.decode("utf-8")

    returncode = process.returncode
    if raises:
        if returncode != 0:
            raise GitCommandError(working_path, cmd)
    else:
        if out.endswith('\n'):
            out = out.strip('\n')
        ui.debug(ui.lightgray, "[%i]" % returncode, ui.reset, out)
        return returncode, out


def get_current_branch(working_path):
    cmd = ("rev-parse", "--abbrev-ref", "HEAD")
    status, output = run_git(working_path, *cmd, raises=False)
    if status != 0:
        raise GitCommandError(working_path, cmd, output=output)

    return output


def get_current_ref(working_path):
    cmd = ("rev-parse", "HEAD")
    status, output = run_git(working_path, *cmd, raises=False)
    if status != 0:
        raise GitCommandError(working_path, cmd, output=output)

    return output


def get_repo_root():
    working_path = path.Path(os.getcwd())
    cmd = ("rev-parse", "--show-toplevel")
    status, output = run_git(working_path, *cmd, raises=False)
    if status != 0:
        raise WorktreeNotFound(working_path)
    return path.Path(output)


def is_dirty(working_path):
    status, _ = run_git(working_path, "diff-index", "--quiet", "HEAD", raises=False)
    return status != 0


def find_ref(repo, candidate_refs):
    """ Find the first reference that exists in the given repo """
    run_git(repo, "fetch", "--all", "--prune")
    for candidate_ref in candidate_refs:
        code, _ = run_git(repo, "rev-parse", candidate_ref, raises=False)
        if code == 0:
            return candidate_ref
    ref_list = ", ".join(candidate_refs)
    raise GitError("Could not find any of:", ref_list, "in repo", repo)


def reset(repo, ref):
    ui.info_2("Resetting", repo, "to", ref)
    run_git(repo, "reset", "--hard", ref)


def get_status(working_path):
    _, out = run_git(working_path, "status", "--porcelain", raises=False)
    for line in out.splitlines():
        if line.startswith("??"):
            return "untracked files"
        if line.startswith(" M"):
            return "modified files"
        if line.startswith("A "):
            return "non-committed files"
    return "clean"



import ruamel.yaml
import pytest

import tsrc.git


class ManifestHandler():
    def __init__(self, path):
        self.path = path
        self.data = {"repos": list()}

    @property
    def yaml_path(self):
        return self.path.joinpath("manifest.yml")

    def init(self):
        to_write = ruamel.yaml.dump(self.data)
        self.yaml_path.write_text(to_write)
        tsrc.git.run_git(self.path, "add", "manifest.yml")
        tsrc.git.run_git(self.path, "commit", "--message", "Add an empty manifest")
        tsrc.git.run_git(self.path, "push", "origin", "master")

    def add_repo(self, src, url):
        self.data["repos"].append({"url": str(url), "src": src})
        self.push(message="add %s" % src)

    def configure_group(self, name, repos):
        groups = self.data.get("groups")
        if not groups:
            self.data["groups"] = dict()
            groups = self.data["groups"]
        groups[name] = dict()
        groups[name]["repos"] = repos
        self.push(message="add %s group" % name)

    def configure_gitlab(self, *, url):
        self.data["gitlab"] = dict()
        self.data["gitlab"]["url"] = url
        self.push("Add gitlab URL: %s" % url)

    def get_repo(self, src):
        for repo in self.data["repos"]:
            if repo["src"] == src:
                return repo
        assert False, "repo '%s' not found in manifest" % src

    def configue_repo(self, src, key, value):
        repo = self.get_repo(src)
        repo[key] = value
        message = "Change %s %s: %s" % (src, key, value)
        self.push(message)

    def set_repo_url(self, src, url):
        self.configue_repo(src, "url", url)

    def set_repo_branch(self, src, branch):
        self.configue_repo(src, "branch", branch)

    def set_repo_ref(self, src, ref):
        self.configue_repo(src, "fixed_ref", ref)

    def set_repo_file_copies(self, src, copies):
        copy_dicts = list()
        for copy_src, copy_dest in copies:
            copy_dicts.append({"src": copy_src, "dest": copy_dest})
        self.configue_repo(src, "copy", copy_dicts)

    def push(self, message):
        to_write = ruamel.yaml.dump(self.data)
        self.yaml_path.write_text(to_write)
        tsrc.git.run_git(self.path, "add", "manifest.yml")
        tsrc.git.run_git(self.path, "commit", "--message", message)
        current_branch = tsrc.git.get_current_branch(self.path)
        tsrc.git.run_git(self.path, "push", "origin", "--set-upstream", current_branch)

    def change_branch(self, branch):
        tsrc.git.run_git(self.path, "checkout", "-B", branch)
        tsrc.git.run_git(self.path, "push", "--no-verify", "origin", "--set-upstream", branch)


class GitServer():
    def __init__(self, tmpdir):
        self.tmpdir = tmpdir
        self.bare_path = tmpdir.joinpath("srv")
        self.src_path = tmpdir.joinpath("src")
        self.add_repo("manifest", add_to_manifest=False)
        self.manifest = ManifestHandler(self.get_path("manifest"))
        self.manifest.init()
        self.manifest_url = self.get_url("manifest")

    def get_path(self, name):
        return self.src_path.joinpath(name)

    def get_url(self, name):
        return str("file://" + self.bare_path.joinpath(name))

    def _create_repo(self, name, empty=False):
        bare_path = self.bare_path.joinpath(name)
        bare_path.makedirs_p()
        tsrc.git.run_git(bare_path, "init", "--bare")
        src_path = self.get_path(name)
        src_path.makedirs_p()
        tsrc.git.run_git(src_path, "init")
        tsrc.git.run_git(src_path, "remote", "add", "origin", bare_path)
        src_path.joinpath("README").touch()
        tsrc.git.run_git(src_path, "add", "README")
        tsrc.git.run_git(src_path, "commit", "--message", "Initial commit")
        if not empty:
            tsrc.git.run_git(src_path, "push", "origin", "master")
        return str(bare_path)

    def add_repo(self, name, add_to_manifest=True, empty=False):
        self._create_repo(name, empty=empty)
        url = self.get_url(name)
        if add_to_manifest:
            self.manifest.add_repo(name, url)
        return url

    def add_group(self, group_name, repos):
        for repo in repos:
            self.add_repo(repo)
        self.manifest.configure_group(group_name, repos)

    def push_file(self, name, file_path, *,
                  contents=None, message=None):
        src_path = self.get_path(name)
        full_path = src_path.joinpath(file_path)
        full_path.parent.makedirs_p()
        full_path.touch()
        if contents:
            full_path.write_text(contents)
        commit_message = message or ("Create/Update %s" % file_path)
        tsrc.git.run_git(src_path, "add", file_path)
        tsrc.git.run_git(src_path, "commit", "--message",
                         commit_message)
        current_branch = tsrc.git.get_current_branch(src_path)
        tsrc.git.run_git(src_path, "push", "origin", "--set-upstream",
                         current_branch)

    def tag(self, name, tag_name):
        src_path = self.get_path(name)
        tsrc.git.run_git(src_path, "tag", tag_name)
        tsrc.git.run_git(src_path, "push", "--no-verify", "origin", tag_name)

    def get_tags(self, name):
        src_path = self.get_path(name)
        rc, out = tsrc.git.run_git(src_path, "tag", raises=False)
        return out

    def get_branches(self, name):
        src_path = self.get_path(name)
        rc, out = tsrc.git.run_git(src_path, "branch", "--list", raises=False)
        return [x[2:].strip() for x in out.splitlines()]

    def change_repo_branch(self, name, new_branch):
        src_path = self.get_path(name)
        tsrc.git.run_git(src_path, "checkout", "-B", new_branch)
        tsrc.git.run_git(src_path, "push", "--no-verify",
                         "origin", "--set-upstream", new_branch)

    def delete_branch(self, name, branch):
        src_path = self.get_path(name)
        tsrc.git.run_git(src_path, "push", "origin", "--delete", branch)


@pytest.fixture
def git_server(tmp_path):
    return GitServer(tmp_path)



""" Custom exceptions """


DOC_URL = "https://tankerapp.github.io/tsrc/ref/formats/"


class Error(Exception):
    """ Base class for our own errors

    """
    def __init__(self, *args):
        super().__init__(self, *args)
        self.message = " ".join(str(x) for x in args)

    def __str__(self):
        return self.message


class InvalidConfig(Error):
    def __init__(self, path, details):
        self.path = path
        self.details = details
        super().__init__(self.detailed_message)

    @property
    def detailed_message(self):
        res = "%s: %s" % (self.path, self.details)
        res += "\n"
        res += "See %s for details" % DOC_URL
        return res

    def __str__(self):
        return self.detailed_message



""" Parse tsrc config files """

import path
import ruamel.yaml
import schema
import xdg

import tsrc


def parse_config_file(file_path, config_schema):
    try:
        contents = file_path.text()
    except OSError as os_error:
        raise tsrc.InvalidConfig(file_path, "Could not read file: %s" % str(os_error))
    try:
        parsed = ruamel.yaml.safe_load(contents)
    except ruamel.yaml.error.YAMLError as yaml_error:
        # pylint: disable=no-member
        context = "(ligne %s, col %s) " % (
            yaml_error.context_mark.line,
            yaml_error.context_mark.column
        )
        message = "%s - YAML error: %s" % (context, yaml_error.context)
        raise tsrc.InvalidConfig(file_path, message)
    try:
        return config_schema.validate(parsed)
    except schema.SchemaError as schema_error:
        raise tsrc.InvalidConfig(file_path, str(schema_error))


def parse_tsrc_config(config_path=None):
    auth_schema = {
        "gitlab": {
            "token": str
        }
    }
    tsrc_schema = schema.Schema({"auth": auth_schema})
    if not config_path:
        config_path = path.Path(xdg.XDG_CONFIG_HOME)
        config_path = config_path.joinpath("tsrc.yml")
    return parse_config_file(config_path, tsrc_schema)



""" Entry point for tsrc log """

import sys

import ui

import tsrc.cli
import tsrc.git


def main(args):
    workspace = tsrc.cli.get_workspace(args)
    workspace.load_manifest()
    all_ok = True
    for unused_index, repo, full_path in workspace.enumerate_repos():
        colors = ["green", "reset", "yellow", "reset", "bold blue", "reset"]
        log_format = "%m {}%h{} - {}%d{} %s {}<%an>{}"
        log_format = log_format.format(*("%C({})".format(x) for x in colors))
        cmd = ["log",
               "--color=always",
               "--pretty=format:%s" % log_format,
               "%s...%s" % (args.from_, args.to)]
        rc, out = tsrc.git.run_git(full_path, *cmd, raises=False)
        if rc != 0:
            all_ok = False
        if out:
            ui.info(ui.bold, repo.src)
            ui.info(ui.bold, "-" * len(repo.src))
            ui.info(out)
    if not all_ok:
        sys.exit(1)



""" Tools to manage tsrc workspaces

Mostly used by tsrc/cli.py

"""

import stat

import ruamel.yaml
import schema
import ui

import tsrc
import tsrc.executor
import tsrc.git
import tsrc.manifest


class LocalManifest:
    """ Represent a manifest that has been cloned locally inside the
    hidden <workspace>/.tsrc directory, along with its configuration

    """
    def __init__(self, workspace_path):
        hidden_path = workspace_path.joinpath(".tsrc")
        self.clone_path = hidden_path.joinpath("manifest")
        self.cfg_path = hidden_path.joinpath("manifest.yml")
        self.manifest = None

    @property
    def branch(self):
        return tsrc.git.get_current_branch(self.clone_path)

    @property
    def copyfiles(self):
        return self.manifest.copyfiles

    @property
    def active_groups(self):
        config = self.load_config()
        return config.get("groups")

    def get_repos(self):
        return self.manifest.get_repos(groups=self.active_groups)

    def load(self):
        yml_path = self.clone_path.joinpath("manifest.yml")
        if not yml_path.exists():
            message = "No manifest found in {}. Did you run `tsrc init` ?"
            raise tsrc.Error(message.format(yml_path))
        self.manifest = tsrc.manifest.load(yml_path)

    def get_gitlab_url(self):
        assert self.manifest, "manifest is empty. Did you call load()?"
        gitlab_config = self.manifest.gitlab
        if not gitlab_config:
            raise tsrc.Error("No gitlab configuration found in manifest")
        return gitlab_config["url"]

    def get_url(self, src):
        return self.manifest.get_url(src)

    def configure(self, url=None, branch="master", tag=None, groups=None):
        if not self.cfg_path.exists() and not url:
            raise tsrc.Error("manifest URL is required when creating a new workspace")
        if self.cfg_path.exists() and not url:
            url = self.load_config()["url"]
        self._ensure_git_state(url, branch=branch, tag=tag)
        self.save_config(url=url, branch=branch, tag=tag, groups=groups)

    def update(self):
        ui.info_2("Updating manifest")
        if not self.clone_path.exists():
            message = "Could not find manifest in {}. "
            message += "Did you run `tsrc init` ?"
            raise tsrc.Error(message.format(self.clone_path))
        cmd = ("fetch", "--prune", "origin")
        tsrc.git.run_git(self.clone_path, *cmd)
        cmd = ("reset", "--hard", "@{u}")
        tsrc.git.run_git(self.clone_path, *cmd)

    def save_config(self, url, branch="master", tag=None, groups=None):
        config = dict()
        config["url"] = url
        config["branch"] = branch
        if tag:
            config["tag"] = tag
        if groups:
            config["groups"] = groups
        with self.cfg_path.open("w") as fp:
            ruamel.yaml.dump(config, fp)

    def load_config(self):
        manifest_schema = schema.Schema({
            "branch": str,
            "url": str,
            schema.Optional("tag"): str,
            schema.Optional("groups"): [str],
        })

        return tsrc.config.parse_config_file(self.cfg_path, manifest_schema)

    def _ensure_git_state(self, url, branch="master", tag=None):
        if self.clone_path.exists():
            tsrc.git.run_git(self.clone_path, "remote", "set-url", "origin", url)

            tsrc.git.run_git(self.clone_path, "fetch")
            tsrc.git.run_git(self.clone_path, "checkout", "-B", branch)
            tsrc.git.run_git(self.clone_path, "branch", branch,
                             "--set-upstream-to", "origin/%s" % branch)
            if tag:
                ref = tag
            else:
                ref = "origin/%s" % branch
            tsrc.git.run_git(self.clone_path, "reset", "--hard", ref)
        else:
            parent, name = self.clone_path.splitpath()
            parent.makedirs_p()
            tsrc.git.run_git(self.clone_path.parent, "clone", url, name, "--branch", branch)
            if tag:
                tsrc.git.run_git(self.clone_path, "reset", "--hard", tag)

    def get_current_branch(self):
        return tsrc.git.get_current_branch(self.clone_path)


class Workspace():
    def __init__(self, root_path):
        self.root_path = root_path
        self.local_manifest = LocalManifest(root_path)

    def joinpath(self, *parts):
        return self.root_path.joinpath(*parts)

    def get_repos(self):
        return self.local_manifest.get_repos()

    def load_manifest(self):
        self.local_manifest.load()

    def get_gitlab_url(self):
        return self.local_manifest.get_gitlab_url()

    def configure_manifest(self, url=None, *, branch="master", tag=None, groups=None):
        self.local_manifest.configure(url=url, branch=branch, tag=tag, groups=groups)

    def update_manifest(self):
        self.local_manifest.update()

    def manifest_branch(self):
        return self.local_manifest.get_current_branch()

    @property
    def active_groups(self):
        return self.local_manifest.active_groups

    def clone_missing(self):
        """ Clone missing repos.

        Called at the beginning of `tsrc init` and `tsrc sync`

        """
        to_clone = list()
        for repo in self.get_repos():
            repo_path = self.joinpath(repo.src)
            if not repo_path.exists():
                to_clone.append(repo)
        cloner = Cloner(self)
        tsrc.executor.run_sequence(to_clone, cloner)

    def set_remotes(self):
        remote_setter = RemoteSetter(self)
        tsrc.executor.run_sequence(self.get_repos(), remote_setter)

    def copy_files(self):
        file_copier = FileCopier(self)
        tsrc.executor.run_sequence(self.local_manifest.copyfiles, file_copier)

    def sync(self):
        syncer = Syncer(self)
        try:
            tsrc.executor.run_sequence(self.get_repos(), syncer)
        finally:
            syncer.display_bad_branches()

    def enumerate_repos(self):
        """ Yield (index, repo, full_path) for all the repos """
        for i, repo in enumerate(self.get_repos()):
            full_path = self.joinpath(repo.src)
            yield (i, repo, full_path)

    def get_url(self, src):
        """ Return the url of the project in `src` """
        return self.local_manifest.get_url(src)


class Cloner(tsrc.executor.Task):
    def __init__(self, workspace):
        self.workspace = workspace

    def description(self):
        return "Cloning missing repos"

    def display_item(self, repo):
        return repo.src

    def process(self, repo):
        ui.info(repo.src)
        repo_path = self.workspace.joinpath(repo.src)
        parent, name = repo_path.splitpath()
        parent.makedirs_p()
        try:
            tsrc.git.run_git(parent, "clone", repo.url, "--branch", repo.branch, name)
        except tsrc.Error:
            raise tsrc.Error("Cloning failed")
        ref = repo.fixed_ref
        if ref:
            ui.info_2("Resetting", repo.src, "to", ref)
            try:
                tsrc.git.run_git(repo_path, "reset", "--hard", ref)
            except tsrc.Error:
                raise tsrc.Error("Resetting to", ref, "failed")


class FileCopier(tsrc.executor.Task):
    def __init__(self, workspace):
        self.workspace = workspace

    def description(self):
        return "Copying files"

    def display_item(self, item):
        src, dest = item
        return "%s -> %s" % (src, dest)

    def process(self, item):
        src, dest = item
        ui.info(src, "->", dest)
        try:
            src_path = self.workspace.joinpath(src)
            dest_path = self.workspace.joinpath(dest)
            if dest_path.exists():
                # Re-set the write permissions on the file:
                dest_path.chmod(stat.S_IWRITE)
            src_path.copy(dest_path)
            # Make sure perms are read only for everyone
            dest_path.chmod(0o10444)
        except Exception as e:
            raise tsrc.Error(str(e))


class RemoteSetter(tsrc.executor.Task):
    def __init__(self, workspace):
        self.workspace = workspace

    def quiet(self):
        return True

    def description(self):
        return "Setting remote URLs"

    def display_item(self, repo):
        return repo.src

    def process(self, repo):
        full_path = self.workspace.joinpath(repo.src)
        try:
            _, old_url = tsrc.git.run_git(full_path, "remote", "get-url", "origin", raises=False)
            if old_url != repo.url:
                ui.info_2(repo.src, old_url, "->", repo.url)
                tsrc.git.run_git(full_path, "remote", "set-url", "origin", repo.url)
        except Exception:
            raise tsrc.Error(repo.src, ":", "Failed to set remote url to %s" % repo.url)


class BadBranches(tsrc.Error):
    pass


class Syncer(tsrc.executor.Task):
    def __init__(self, workspace):
        self.workspace = workspace
        self.bad_branches = list()

    def description(self):
        return "Synchronize workspace"

    def display_item(self, repo):
        return repo.src

    def process(self, repo):
        ui.info(repo.src)
        repo_path = self.workspace.joinpath(repo.src)
        self.check_branch(repo, repo_path)
        self.fetch(repo_path)

        if repo.fixed_ref:
            self.sync_repo_to_ref(repo_path, repo.fixed_ref)
        else:
            self.sync_repo_to_branch(repo_path)

    def check_branch(self, repo, repo_path):
        current_branch = None
        try:
            current_branch = tsrc.git.get_current_branch(repo_path)
        except tsrc.Error:
            raise tsrc.Error("Not on any branch")

        if current_branch and current_branch != repo.branch:
            self.bad_branches.append((repo.src, current_branch, repo.branch))

    @staticmethod
    def fetch(repo_path):
        try:
            tsrc.git.run_git(repo_path, "fetch", "--tags", "--prune", "origin")
        except tsrc.Error:
            raise tsrc.Error("fetch failed")

    @staticmethod
    def sync_repo_to_ref(repo_path, ref):
        ui.info_2("Resetting to", ref)
        status = tsrc.git.get_status(repo_path)
        if status != "clean":
            raise tsrc.Error("%s, skipping" % status)
        try:
            tsrc.git.run_git(repo_path, "reset", "--hard", ref)
        except tsrc.Error:
            raise tsrc.Error("updating ref failed")

    @staticmethod
    def sync_repo_to_branch(repo_path):
        try:
            tsrc.git.run_git(repo_path, "merge", "--ff-only", "@{u}")
        except tsrc.Error:
            raise tsrc.Error("updating branch failed")

    def display_bad_branches(self):
        if not self.bad_branches:
            return
        ui.error("Some projects were not on the correct branch")
        headers = ("project", "actual", "expected")
        data = [
            ((ui.bold, name), (ui.red, actual), (ui.green, expected)) for
            (name, actual, expected) in self.bad_branches
        ]
        ui.info_table(data, headers=headers)
        raise BadBranches()



""" Repo value object """

import attr


# pylint: disable=too-few-public-methods
@attr.s(frozen=True)
class Repo():
    src = attr.ib()
    url = attr.ib()
    branch = attr.ib(default="master")
    fixed_ref = attr.ib(default=None)



import os.path

import ruamel.yaml

import tsrc.manifest
from tsrc.repo import Repo

import pytest


def test_load():
    contents = """
gitlab:
  url: http://gitlab.example.com
repos:
  - src: foo
    url: git@example.com:foo.git
    branch: next

  - src: master
    url: git@example.com:master.git
    fixed_ref: v0.1
    copy:
      - src: top.cmake
        dest: CMakeLists.txt
      - src: .clang-format
"""
    manifest = tsrc.manifest.Manifest()
    parsed = ruamel.yaml.safe_load(contents)
    manifest.load(parsed)
    assert manifest.gitlab["url"] == "http://gitlab.example.com"
    assert manifest.get_repos() == [
        tsrc.Repo(
            url="git@example.com:foo.git",
            src="foo",
            branch="next",
            fixed_ref=None,
        ),
        tsrc.Repo(
            url="git@example.com:master.git",
            src="master",
            branch="master",
            fixed_ref="v0.1"
        ),
    ]
    assert manifest.copyfiles == [
        (os.path.join("master", "top.cmake"), "CMakeLists.txt"),
        (os.path.join("master", ".clang-format"), ".clang-format"),
    ]


def test_find():
    contents = """
repos:
  - src: foo
    url: git@example.com:proj_one/foo

  - src: bar
    url: git@example.com:proj_two/bar
"""
    manifest = tsrc.manifest.Manifest()
    parsed = ruamel.yaml.safe_load(contents)
    manifest.load(parsed)
    assert manifest.get_url("foo") == "git@example.com:proj_one/foo"
    assert manifest.get_url("bar") == "git@example.com:proj_two/bar"
    with pytest.raises(tsrc.manifest.RepoNotFound) as e:
        manifest.get_url("no/such")
        assert "no/such" in e.value.message


def test_validates(tmp_path):
    contents = """
repos:
  - src: bar
    url: baz
    copy:
      - src: foo
        dest: bar
gitlab:
  url: foo
"""
    manifest_path = tmp_path.joinpath("manifest.yml")
    manifest_path.write_text(contents)
    res = tsrc.manifest.load(manifest_path)
    assert res


class ReposGetter:
    def __init__(self, tmp_path):
        self.tmp_path = tmp_path
        self.contents = None

    def get_repos(self, groups=None, all_=None):
        manifest_path = self.tmp_path.joinpath("manifest.yml")
        manifest_path.write_text(self.contents)
        manifest = tsrc.manifest.load(manifest_path)
        return [repo.src for repo in manifest.get_repos(groups=groups, all_=all_)]


@pytest.fixture
def repos_getter(tmp_path):
    return ReposGetter(tmp_path)


def test_default_group(repos_getter):
    contents = """
repos:
  - { src: one, url: one.com }
  - { src: two, url: two.com }
  - { src: three, url: three.com }

groups:
  default:
    repos: [one, two]
"""
    repos_getter.contents = contents
    assert repos_getter.get_repos(groups=None) == ["one", "two"]


def test_specific_group(repos_getter):
    contents = """
repos:
  - { src: any, url: any.com }
  - { src: linux1, url: linux1.com }
  - { src: linux2, url: linux2.com }

groups:
  default:
    repos: [any]
  linux:
    repos: [linux1, linux2]
"""
    repos_getter.contents = contents
    assert repos_getter.get_repos(groups=["default", "linux"]) == ["any", "linux1", "linux2"]


def test_inclusion(repos_getter):
    contents = """
repos:
  - { src: a, url: a.com }
  - { src: b, url: b.com }
  - { src: c, url: c.com }

groups:
  a_group:
    repos: [a]
  b_group:
     repos: [b]
     includes: [a_group]
  c_group:
      repos: [c]
      includes: [b_group]
"""
    repos_getter.contents = contents
    assert repos_getter.get_repos(groups=["c_group"]) == ["a", "b", "c"]


def test_all_repos(repos_getter):
    contents = """
repos:
  - { src: one, url: one.com }
  - { src: two, url: two.com }

groups:
  default:
    repos: [one]
"""
    repos_getter.contents = contents
    assert repos_getter.get_repos(all_=False) == ["one"]
    assert repos_getter.get_repos(all_=True) == ["one", "two"]



import types
import mock

import pytest

import tsrc.cli.push
import tsrc.git
import tsrc.gitlab

GITLAB_URL = "http://gitlab.example.com"
TIMOTHEE = {"name": "timothee", "id": 1}
THEO = {"name": "theo", "id": 2}
JOHN = {"name": "john", "id": 3}
BART = {"name": "bart", "id": 4}

MR_STUB = {"id": "3978", "iid": "42", "web_url": "http://mr/42", "title": "Boring title"}

PROJECT_IDS = {
    "foo/bar": "42"
}


@pytest.fixture
def foo_path(monkeypatch, git_server, tsrc_cli, workspace_path):
    """ Path to a freshly cloned repository """
    git_server.manifest.configure_gitlab(url=GITLAB_URL)
    git_server.add_repo("foo/bar")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    foo_path = workspace_path.joinpath("foo/bar")
    monkeypatch.chdir(foo_path)


@pytest.fixture
def push_args():
    args = types.SimpleNamespace()
    args.accept = False
    args.target_branch = "master"
    args.mr_title = None
    args.assignee = None
    args.force = False
    args.ready = None
    args.wip = None
    return args


@pytest.fixture
def gitlab_mock():
    all_users = [JOHN, BART, TIMOTHEE, THEO]

    def get_project_members(project_id, query):
        assert project_id in PROJECT_IDS.values()
        return [user for user in all_users if query in user["name"]]

    def get_group_members(group_name, query):
        return [user for user in all_users if query in user["name"]]

    gl_mock = mock.create_autospec(tsrc.gitlab.GitLabHelper, instance=True)
    gl_mock.get_project_members = get_project_members
    gl_mock.get_group_members = get_group_members
    gl_mock.get_project_id = lambda x: PROJECT_IDS[x]
    # Define a few helper methods to make tests nicer to read:
    new_defs = {
        "assert_mr_created": gl_mock.create_merge_request.assert_called_with,
        "assert_mr_not_created": gl_mock.create_merge_request.assert_not_called,
        "assert_mr_updated": gl_mock.update_merge_request.assert_called_with,
        "assert_mr_accepted": gl_mock.accept_merge_request.assert_called_with,
    }
    for name, func in new_defs.items():
        setattr(gl_mock, name, func)
    return gl_mock


def test_creating_merge_request(foo_path, tsrc_cli, gitlab_mock, push_args):
    tsrc.git.run_git(foo_path, "checkout", "-b", "new-feature")
    tsrc.git.run_git(foo_path, "commit", "--message", "new feature", "--allow-empty")

    gitlab_mock.find_opened_merge_request.return_value = list()
    gitlab_mock.create_merge_request.return_value = MR_STUB

    push_args.assignee = "john"
    push_args.target_branch = "next"
    push_args.mr_title = "Best feature ever"

    push_action = tsrc.cli.push.PushAction(push_args, gl_helper=gitlab_mock)

    push_action.main()

    gitlab_mock.assert_mr_created(
        "42", "new-feature",
        target_branch="next",
        title="new-feature"
    )
    gitlab_mock.assert_mr_updated(
        MR_STUB,
        assignee_id=JOHN["id"],
        remove_source_branch=True,
        target_branch="next",
        title="Best feature ever"
    )


def test_existing_merge_request(foo_path, tsrc_cli, gitlab_mock, push_args):
    tsrc.git.run_git(foo_path, "checkout", "-b", "new-feature")
    tsrc.git.run_git(foo_path, "commit", "--message", "new feature", "--allow-empty")

    gitlab_mock.find_opened_merge_request.return_value = MR_STUB

    push_args.target_branch = "next"
    push_args.mr_title = "Best feature ever"
    push_action = tsrc.cli.push.PushAction(push_args, gl_helper=gitlab_mock)
    push_action.main()

    gitlab_mock.assert_mr_not_created()
    gitlab_mock.assert_mr_updated(
        MR_STUB,
        remove_source_branch=True,
        target_branch="next",
        title="Best feature ever"
    )


def test_accept_merge_request(foo_path, tsrc_cli, gitlab_mock, push_args):
    tsrc.git.run_git(foo_path, "checkout", "-b", "new-feature")
    tsrc.git.run_git(foo_path, "commit", "--message", "new feature", "--allow-empty")

    gitlab_mock.find_opened_merge_request.return_value = MR_STUB

    push_args.accept = True
    push_action = tsrc.cli.push.PushAction(push_args, gl_helper=gitlab_mock)
    push_action.main()

    gitlab_mock.assert_mr_accepted(MR_STUB)


def test_unwipify_existing_merge_request(foo_path, tsrc_cli, gitlab_mock, push_args):
    existing_mr = {
        "title": "WIP: nice title",
        "web_url": "http://example.com/42",
        "iid": "42",
    }
    gitlab_mock.find_opened_merge_request.return_value = existing_mr

    push_args.ready = True
    push_action = tsrc.cli.push.PushAction(push_args, gl_helper=gitlab_mock)
    push_action.main()

    gitlab_mock.assert_mr_not_created()
    gitlab_mock.assert_mr_updated(
        existing_mr,
        remove_source_branch=True,
        target_branch="master",
        title="nice title"
    )



import os

import pytest
import ruamel

import tsrc.cli


def repo_exists(workspace_path, repo):
    return workspace_path.joinpath(repo).exists()


def assert_cloned(workspace_path, repo):
    assert repo_exists(workspace_path, repo)


def assert_not_cloned(workspace_path, repo):
    assert not repo_exists(workspace_path, repo)


def test_init(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("foo/bar")
    git_server.add_repo("spam/eggs")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    assert_cloned(workspace_path, "foo/bar")
    assert_cloned(workspace_path, "spam/eggs")


def test_init_with_args(tsrc_cli, git_server, monkeypatch, tmp_path):
    git_server.add_repo("foo")
    work2_path = tmp_path.joinpath("work2").mkdir()
    tsrc_cli.run("init", "--workspace", work2_path, git_server.manifest_url)
    assert_cloned(work2_path, "foo")


def test_init_twice(tsrc_cli, git_server):
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    tsrc_cli.run("init", manifest_url)


def test_init_maint_manifest_branch(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("bar")
    # foo repo will only exist on the 'devel' branch of the manifest:
    git_server.manifest.change_branch("devel")
    git_server.add_repo("foo")

    tsrc_cli.run("init", "--branch", "devel", git_server.manifest_url)

    assert_cloned(workspace_path, "foo")


def test_change_repo_url(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("foo")
    tsrc_cli.run("init", git_server.manifest_url)
    new_url = "git@example.com/foo"
    git_server.manifest.set_repo_url("foo", new_url)
    tsrc_cli.run("init", git_server.manifest_url)
    assert_cloned(workspace_path, "foo")
    foo_path = workspace_path.joinpath("foo")
    _, actual_url = tsrc.git.run_git(foo_path, "remote", "get-url", "origin", raises=False)
    assert actual_url == new_url


def test_copy_files(tsrc_cli, git_server, workspace_path):
    manifest_url = git_server.manifest_url
    git_server.add_repo("master")
    top_cmake_contents = "# Top CMakeLists.txt"
    git_server.push_file("master", "top.cmake", contents=top_cmake_contents)
    git_server.manifest.set_repo_file_copies("master", [["top.cmake", "CMakeLists.txt"]])

    tsrc_cli.run("init", manifest_url)

    assert workspace_path.joinpath("CMakeLists.txt").text() == top_cmake_contents


def test_uses_correct_branch_for_repo(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("foo")
    git_server.change_repo_branch("foo", "next")
    git_server.push_file("foo", "next.txt")
    git_server.manifest.set_repo_branch("foo", "next")

    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)

    foo_path = workspace_path.joinpath("foo")
    assert tsrc.git.get_current_branch(foo_path) == "next"


def test_empty_repo(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("foo", empty=True)
    git_server.add_repo("bar")

    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url, expect_fail=True)


def test_resets_to_fixed_ref(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("foo")
    git_server.tag("foo", "v1.0")
    git_server.push_file("foo", "2.txt", message="Working on v2")
    git_server.manifest.set_repo_ref("foo", "v1.0")

    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)

    foo_path = workspace_path.joinpath("foo")
    expected_ref = tsrc.git.run_git(foo_path, "rev-parse", "v1.0", raises=False)
    actual_ref = tsrc.git.run_git(foo_path, "rev-parse", "HEAD", raises=False)
    assert expected_ref == actual_ref


def test_use_default_group(tsrc_cli, git_server, workspace_path):
    git_server.add_group("default", ["a", "b"])
    git_server.add_repo("c")

    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)

    assert_cloned(workspace_path, "a")
    assert_not_cloned(workspace_path, "c")


def test_use_specific_group(tsrc_cli, git_server, workspace_path):
    git_server.add_group("foo", ["bar", "baz"])
    git_server.add_group("spam", ["eggs", "beacon"])
    git_server.add_repo("other")

    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url, "-g", "foo", "-g", "spam")

    assert_cloned(workspace_path, "bar")
    assert_cloned(workspace_path, "eggs")
    assert_not_cloned(workspace_path, "other")


def test_change_branch(tsrc_cli, git_server, workspace_path):
    git_server.add_repo("one")
    git_server.manifest.change_branch("next")
    git_server.add_repo("two")

    tsrc_cli.run("init", git_server.manifest_url)
    assert_not_cloned(workspace_path, "two")

    tsrc_cli.run("init", "--branch", "next")
    assert_cloned(workspace_path, "two")



""" Helpers to run things on multiple repos and collect errors """

import abc

import ui

import tsrc


class ExecutorFailed(tsrc.Error):
    pass


class Task(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def description(self) -> str:
        pass

    # pylint: disable=no-self-use
    def quiet(self):
        return False

    @abc.abstractmethod
    def display_item(self, _) -> str:
        pass

    @abc.abstractmethod
    def process(self, _):
        pass


class SequentialExecutor():
    def __init__(self, task):
        self.task = task
        self.errors = list()

    def process(self, items):
        if not items:
            return True
        ui.info_1(self.task.description())

        self.errors = list()
        num_items = len(items)
        for i, item in enumerate(items):
            if not self.task.quiet():
                ui.info_count(i, num_items, end="")
            self.process_one(item)

        if self.errors:
            self.handle_errors()

    def handle_errors(self):
        ui.error(self.task.description(), "failed")
        for item, error in self.errors:
            item_desc = self.task.display_item(item)
            message = [ui.green, "*", " ", ui.reset, ui.bold, item_desc]
            if error.message:
                message.extend([ui.reset, ": ", error.message])
            ui.info(*message, sep="")
        raise ExecutorFailed()

    def process_one(self, item):
        try:
            self.task.process(item)
        except tsrc.Error as error:
            self.errors.append((item, error))


def run_sequence(items, task):
    executor = SequentialExecutor(task)
    return executor.process(items)



import tsrc.groups

import pytest


def test_happy_grouping():
    group_list = tsrc.groups.GroupList(elements={"a", "b", "c"})
    group_list.add("default", {"a", "b"})
    group_list.add("other", {"c"}, includes={"default"})
    actual = group_list.get_elements(groups={"other"})
    assert actual == {"a", "b", "c"}


def test_default_is_all():
    group_list = tsrc.groups.GroupList(elements={"a", "b", "c"})
    assert group_list.get_elements() == {"a", "b", "c"}


def test_unknown_element():
    group_list = tsrc.groups.GroupList(elements={"a", "b", "c"})
    with pytest.raises(tsrc.groups.UnknownElement) as e:
        group_list.add("invalid-group", {"no-such-element"})
    assert e.value.group_name == "invalid-group"
    assert e.value.element == "no-such-element"


def test_unknown_include():
    group_list = tsrc.groups.GroupList(elements={"a", "b", "c"})
    group_list.add("default", {"a", "b"})
    group_list.add("invalid-group", {"c"}, includes={"no-such-group"})
    with pytest.raises(tsrc.groups.GroupNotFound) as e:
        group_list.get_elements(groups={"invalid-group"})
    assert e.value.parent_group.name == "invalid-group"
    assert e.value.group_name == "no-such-group"


def test_diamond():
    group_list = tsrc.groups.GroupList(elements={"a", "b", "c", "d"})
    group_list.add("top", {"a"})
    group_list.add("left", {"b"}, includes=["top"])
    group_list.add("right", {"c"}, includes=["top"])
    group_list.add("bottom", {"d"}, includes=["left", "right"])
    actual = group_list.get_elements(groups=["bottom"])
    assert actual == {"a", "b", "c", "d"}


def test_ping_pong():
    group_list = tsrc.groups.GroupList(elements={"a", "b"})
    group_list.add("ping", {"a"}, includes=["pong"])
    group_list.add("pong", {"b"}, includes=["ping"])
    actual = group_list.get_elements(groups=["ping"])
    assert actual == {"a", "b"}


def test_circle():
    group_list = tsrc.groups.GroupList(elements={"a", "b", "c"})
    group_list.add("a", {"a"}, includes=["b"])
    group_list.add("b", {"b"}, includes=["c"])
    group_list.add("c", {"c"}, includes=["a"])
    actual = group_list.get_elements(groups=["a"])
    assert actual == {"a", "b", "c"}


def test_unknown_group():
    group_list = tsrc.groups.GroupList(elements={"a", "b", "c"})
    group_list.add("default", {"a", "b"})
    with pytest.raises(tsrc.groups.GroupNotFound) as e:
        group_list.get_elements(groups={"no-such-group"})
    assert e.value.parent_group is None
    assert e.value.group_name == "no-such-group"



""" Main tsrc entry point """

import argparse
import functools
import importlib
import sys
import textwrap

import colored_traceback
import ui

import tsrc


def fix_cmd_args_for_foreach(args, foreach_parser):
    """ We want to support both:
      $ tsrc foreach -c 'shell command'
     and
      $ tsrc foreach -- some-cmd --some-opts

    Due to argparse limitations, args.cmd will always be
    a list, but we nee a *string* when using 'shell=True'

    So transform the argparse.Namespace object to have
    * args.cmd suitable  to pass to subprocess later
    * args.cmd_as_str suitable for display purposes

    """
    def die(message):
        ui.error(message)
        print(foreach_parser.epilog, end="")
        sys.exit(1)

    if args.shell:
        if len(args.cmd) != 1:
            die("foreach -c must be followed by exactly one argument")
        cmd = args.cmd[0]
        cmd_as_str = cmd
    else:
        cmd = args.cmd
        if not cmd:
            die("needs a command to run")
        cmd_as_str = " ".join(cmd)

    args.cmd = cmd
    args.cmd_as_str = cmd_as_str


def workspace_subparser(subparser, name):
    parser = subparser.add_parser(name)
    parser.add_argument("-w", "--workspace", dest="workspace_path")
    return parser


def main_wrapper(main_func):
    """ Wraps main() entry point to better deal with errors """
    @functools.wraps(main_func)
    def wrapped(args=None):
        colored_traceback.add_hook()
        try:
            main_func(args=args)
        except tsrc.Error as e:
            # "expected" failure, display it and exit
            if e.message:
                ui.error(e.message)
            sys.exit(1)
        except KeyboardInterrupt:
            ui.warning("Interrupted by user, quitting")
            sys.exit(1)
    return wrapped


@main_wrapper
def main(args=None):
    parser = argparse.ArgumentParser()

    parser.add_argument("--verbose", help="Show debug messages",
                        action="store_true")
    parser.add_argument("-q", "--quiet", help="Only display warnings and errors",
                        action="store_true")
    parser.add_argument("--color", choices=["auto", "always", "never"])

    subparsers = parser.add_subparsers(title="subcommands", dest="command")

    subparsers.add_parser("version")

    foreach_parser = workspace_subparser(subparsers, "foreach")
    foreach_parser.add_argument("cmd", nargs="*")
    foreach_parser.add_argument("-c", dest="shell", action="store_true")
    foreach_parser.epilog = textwrap.dedent("""\
    Usage:
       # Run command directly
       tsrc foreach -- some-cmd --with-option
    Or:
       # Run command through the shell
       tsrc foreach -c 'some cmd'
    """)
    foreach_parser.formatter_class = argparse.RawDescriptionHelpFormatter

    init_parser = workspace_subparser(subparsers, "init")
    init_parser.add_argument("manifest_url", nargs="?")
    init_parser.add_argument("-b", "--branch")
    init_parser.add_argument("-g", "--group", action="append", dest="groups")
    init_parser.set_defaults(branch="master")

    log_parser = workspace_subparser(subparsers, "log")
    log_parser.add_argument("--from", required=True, dest="from_", metavar="FROM")
    log_parser.add_argument("--to")
    log_parser.set_defaults(to="HEAD")

    push_parser = workspace_subparser(subparsers, "push")
    push_parser.add_argument("--accept", action="store_true", default=False)
    push_parser.add_argument("-f", "--force", action="store_true", default=False)
    push_parser.add_argument("-t", "--target", dest="target_branch", default="master")
    message_group = push_parser.add_mutually_exclusive_group()
    message_group.add_argument("-m", "--message", dest="mr_title")
    message_group.add_argument("--wip", action="store_true", help="Mark merge request as WIP")
    message_group.add_argument("--ready", action="store_true", help="Mark merge request as ready")
    push_parser.add_argument("-a", "--assignee", dest="assignee")
    workspace_subparser(subparsers, "status")
    workspace_subparser(subparsers, "sync")

    args = parser.parse_args(args=args)
    ui.setup(verbose=args.verbose, quiet=args.quiet, color=args.color)

    command = args.command
    if not command:
        parser.print_help()
        sys.exit(1)
    module = importlib.import_module("tsrc.cli.%s" % command)
    if command == "foreach":
        fix_cmd_args_for_foreach(args, foreach_parser)

    return module.main(args)



""" Entry point for tsrc status """

import attr
import ui

import tsrc.cli


# pylint: disable=too-few-public-methods
@attr.s
class Status:
    src = attr.ib()
    branch = attr.ib()
    dirty = attr.ib()


def collect_statuses(workspace):
    errors = list()
    result = list()
    repos = workspace.get_repos()

    if not repos:
        return errors, result

    num_repos = len(repos)
    max_len = max((len(x.src) for x in repos))
    for i, repo, full_path in workspace.enumerate_repos():
        ui.info_count(i, num_repos,
                      "Checking", repo.src.ljust(max_len + 1), end="\r")
        try:
            branch = tsrc.git.get_current_branch(full_path)
        except tsrc.git.GitError as e:
            errors.append((repo.src, e))
            continue

        dirty = tsrc.git.is_dirty(full_path)

        result.append(Status(src=repo.src, branch=branch, dirty=dirty))

    ui.info("")
    return result, errors


def display_statuses(statuses, errors):
    if not statuses:
        return
    max_src = max((len(x.src) for x in statuses))
    for status in statuses:
        message = (ui.green, "*", ui.reset, ui.bold, status.src.ljust(max_src),
                   ui.reset, ui.green, status.branch)
        if status.dirty:
            message = message + (ui.reset, ui.brown, "(dirty)")
        ui.info(*message)

    if errors:
        ui.info()
        ui.error("Errors when getting branch")
        for src, error in errors:
            ui.info("*", ui.bold, src, ui.reset, error.output)
        ui.info()


def main(args):
    workspace = tsrc.cli.get_workspace(args)
    workspace.load_manifest()
    statuses, errors = collect_statuses(workspace)
    display_statuses(statuses, errors)



""" Fixtures for tsrc testing """

import os
import re

import path
import pytest
import ruamel.yaml
import ui

import tsrc.cli.main
import tsrc.git
import tsrc.workspace

from ui.tests.conftest import message_recorder
from .helpers.git_server import git_server
from .helpers.cli import tsrc_cli


@pytest.fixture()
def tmp_path(tmpdir):
    """ Convert py.path.Local() to path.Path() objects """
    return path.Path(tmpdir.strpath)


@pytest.fixture
def workspace_path(tmp_path):
    return tmp_path.joinpath("work").mkdir()


@pytest.fixture
def workspace(workspace_path):
    return tsrc.workspace.Workspace(workspace_path)



import os

import pytest

import tsrc.cli

from tsrc.test.conftest import *


def test_status_happy(tsrc_cli, git_server, workspace_path, message_recorder):
    git_server.add_repo("foo/bar")
    git_server.add_repo("spam/eggs")
    git_server.push_file("foo/bar", "CMakeLists.txt")
    git_server.push_file("spam/eggs", "CMakeLists.txt")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    tsrc.git.run_git(workspace_path.joinpath("spam", "eggs"),
                     "checkout", "-b", "fish")

    tsrc_cli.run("status")

    assert message_recorder.find("\* foo/bar   master")
    assert message_recorder.find("\* spam/eggs fish")


def test_status_dirty(tsrc_cli, git_server, workspace_path, message_recorder):
    git_server.add_repo("foo/bar")
    git_server.push_file("foo/bar", "CMakeLists.txt")
    manifest_url = git_server.manifest_url
    tsrc_cli.run("init", manifest_url)
    workspace_path.joinpath("foo", "bar", "CMakeLists.txt").write_text("DIRTY FILE")

    tsrc_cli.run("status")

    assert message_recorder.find("\* foo/bar master \(dirty\)")


def test_status_error(tsrc_cli, git_server, workspace_path, message_recorder):
    git_server.add_repo("foo/bar")
    git_server.add_repo("spam/eggs")
    git_server.push_file("foo/bar", "CMakeLists.txt")
    git_server.push_file("spam/eggs", "CMakeLists.txt")
    tsrc_cli.run("init", git_server.manifest_url)
    # corrupt the git
    workspace_path.joinpath("spam", "eggs", ".git", "HEAD").remove()

    tsrc_cli.run("status")

    assert message_recorder.find("\* foo/bar master")
    assert message_recorder.find("Errors when getting branch")



import pytest
import ui

import tsrc
import tsrc.executor


class Kaboom(tsrc.Error):
    def __str__(self):
        return "Kaboom!"


class FakeTask(tsrc.executor.Task):
    def __init__(self):
        pass

    def description(self):
        print("Frobnicating all items")

    def display_item(self, item):
        return item

    def process(self, item):
        ui.info("frobnicate", item)
        if item == "bar":
            print("ko :/")
            raise Kaboom()
        ui.info("ok !")


def test_doing_nothing():
    task = FakeTask()
    tsrc.executor.run_sequence(list(), task)


def test_happy():
    task = FakeTask()
    tsrc.executor.run_sequence(["foo", "spam"], task)


def test_collect_errors():
    task = FakeTask()
    with pytest.raises(tsrc.executor.ExecutorFailed):
        tsrc.executor.run_sequence(["foo", "bar"], task)



""" Entry point for tsrc version """

import pkg_resources

import path
import ui

import tsrc.git


# pylint: disable=unused-argument
def main(args):
    tsrc_distribution = pkg_resources.get_distribution("tsrc")
    # pylint: disable=no-member
    version = tsrc_distribution.version
    location = path.Path(tsrc_distribution.location)
    dirty = False
    short_hash = None
    rc, out = tsrc.git.run_git(location, "rev-parse", "--short", "HEAD", raises=False)
    if rc == 0:
        short_hash = out
        dirty = tsrc.git.is_dirty(location)
    message = "tsrc version %s" % version
    if short_hash:
        message += " - git: %s" % short_hash
        if dirty:
            message += " (dirty)"
    ui.info(message)



import sys
from setuptools import setup, find_packages

if sys.version_info.major < 3:
    sys.exit("Error: Please upgrade to Python3")


def get_long_description():
    with open("README.rst") as fp:
        return fp.read()


setup(name="tsrc",
      version="0.3.0",
      description="Manage multiple repositories",
      long_description=get_long_description(),
      url="https://github.com/TankerApp/tsrc",
      author="Kontrol SAS",
      packages=find_packages(),
      include_package_data=True,
      install_requires=[
        "attrs",
        "colored_traceback",
        "colorama",
        "path.py",
        "python-cli-ui",
        "pyparsing",
        "requests",
        "ruamel.yaml",
        "schema",
        "tabulate",
        "unidecode",
        "xdg",
      ],
      classifiers=[
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
      ],
      entry_points={
        "console_scripts": [
          "tsrc = tsrc.cli.main:main",
         ]
      }
      )



""" Entry point for `tsrc init` """
import os

import path
import ui

import tsrc.workspace


def main(args):
    workspace_path = args.workspace_path or os.getcwd()
    workspace = tsrc.workspace.Workspace(path.Path(workspace_path))
    ui.info_1("Configuring workspace in", ui.bold, workspace_path)
    workspace.configure_manifest(url=args.manifest_url, branch=args.branch, groups=args.groups)
    workspace.load_manifest()
    workspace.clone_missing()
    workspace.set_remotes()
    workspace.copy_files()
    ui.info("Done", ui.check)



import mock

import pytest

import tsrc.git
import tsrc.gitlab
import tsrc.manifest


def test_tsrc_cli_help(tsrc_cli):
    tsrc_cli.run("--help")


def test_tsrc_cli_bad_args(tsrc_cli):
    tsrc_cli.run("bad", expect_fail=True)


def read_remote_manifest(workspace_path, git_server):
    tsrc.git.run_git(workspace_path, "clone", git_server.manifest_url)
    manifest_yml = workspace_path.joinpath("manifest", "manifest.yml")
    manifest = tsrc.manifest.load(manifest_yml)
    return manifest


def test_git_server_add_repo_can_clone(workspace_path, git_server):
    foobar_url = git_server.add_repo("foo/bar")
    tsrc.git.run_git(workspace_path, "clone", foobar_url)
    assert workspace_path.joinpath("bar").exists()


def test_git_server_can_add_copies(workspace_path, git_server):
    git_server.add_repo("foo")
    git_server.manifest.set_repo_file_copies("foo", [["foo.txt", "top.txt"]])
    manifest = read_remote_manifest(workspace_path, git_server)
    assert manifest.copyfiles == [("foo/foo.txt", "top.txt")]


def test_can_configure_gitlab(tmp_path, git_server):
    test_url = "http://gitlab.example.org"
    git_server.manifest.configure_gitlab(url=test_url)
    manifest = read_remote_manifest(tmp_path, git_server)
    assert manifest.gitlab["url"] == test_url


def test_git_server_add_repo_updates_manifest(workspace_path, git_server):
    git_server.add_repo("foo/bar")
    git_server.add_repo("spam/eggs")
    manifest = read_remote_manifest(workspace_path, git_server)
    repos = manifest.get_repos()
    assert len(repos) == 2
    for repo in repos:
        rc, out = tsrc.git.run_git(workspace_path, "ls-remote", repo.url, raises=False)
        assert rc == 0
        assert "refs/heads/master" in out


def test_git_server_change_manifest_branch(workspace_path, git_server):
    git_server.add_repo("foo")
    git_server.manifest.change_branch("devel")
    git_server.add_repo("bar")

    tsrc.git.run_git(workspace_path,
                     "clone", git_server.manifest_url, "--branch", "devel")
    manifest_yml = workspace_path.joinpath("manifest", "manifest.yml")
    manifest = tsrc.manifest.load(manifest_yml)

    assert len(manifest.get_repos()) == 2


def test_git_server_change_repo_branch(workspace_path, git_server):
    foo_url = git_server.add_repo("foo")
    git_server.change_repo_branch("foo", "devel")
    git_server.push_file("foo", "devel.txt", contents="this is devel\n")
    tsrc.git.run_git(workspace_path, "clone", foo_url, "--branch", "devel")
    foo_path = workspace_path.joinpath("foo")
    assert foo_path.joinpath("devel.txt").text() == "this is devel\n"


def test_git_server_tag(workspace_path, git_server):
    foo_url = git_server.add_repo("foo")
    git_server.tag("foo", "v0.1")
    rc, out = tsrc.git.run_git(workspace_path, "ls-remote", foo_url, raises=False)
    assert rc == 0
    assert "refs/tags/v0.1" in out
